"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

import click  # type: ignore
from ifpd2 import __version__
from ifpd2.const import CONTEXT_SETTINGS
from ifpd2.scripts import db, extract_kmers, query, query2
import sys
import webbrowser


@click.group(
    name="ifpd2",
    context_settings=CONTEXT_SETTINGS,
    help=f"""\b
Version:    {__version__}
Author:     Gabriele Girelli
Docs:       http://ggirelli.github.io/ifpd2
Code:       http://github.com/ggirelli/ifpd2

Another iFISH probe design pipeline (II).""",
)
@click.version_option(__version__)
def main():
    """Just a hook for the entry point. Silence is golden!"""


@click.command(
    "_docs",
    help="Open online documentation on your favorite browser.",
)
def open_documentation() -> None:
    webbrowser.open("https://ggirelli.github.io/radiantkit/")
    sys.exit()


main.add_command(open_documentation)
main.add_command(extract_kmers.main)
main.add_command(db.run.main)
main.add_command(query.main)
main.add_command(query2.main)



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

from ifpd2 import asserts as ass
from ifpd2.dataclasses import GCRange
import logging
import numpy as np  # type: ignore
import pandas as pd  # type: ignore
from tqdm import tqdm  # type: ignore
from typing import List, Tuple


class Oligo:
    """Oligo database line/record parser.
    Presents oligo values as properties.
    """

    _data = None

    __colnames = [
        "name",
        "chrom",
        "start",
        "end",
        "tm_dG",
        "dH",
        "dS",
        "Tm",
        "seq",
        "nOT",
        "k",
        "ss_dG",
        "GC",
    ]

    def __init__(self, oligo, i):
        super(Oligo, self).__init__()
        ass.ert_type(i, int, "oligo id")
        if i < 0:
            raise AssertionError
        self._raw_data = oligo.strip().split("\t")
        for i in [2, 3, 9, 10]:
            self._raw_data[i] = int(self._raw_data[i])
        for i in [4, 5, 6, 7, 11, 12]:
            self._raw_data[i] = float(self._raw_data[i])
        self._idx = i

    @property
    def idx(self):
        return self._idx

    @property
    def raw_data(self):
        return self._raw_data

    @property
    def data(self):
        if isinstance(self._data, type(None)):
            self._data = pd.DataFrame(
                self._raw_data, columns=[self._idx], index=self.__colnames
            ).transpose()
        return self._data

    @property
    def start(self):
        return self.data["start"].values[0]

    @property
    def end(self):
        return self.data["end"].values[0]

    @property
    def score(self):
        if "score" in self.data.columns:
            return self.data["score"].values[0]
        return np.nan

    @staticmethod
    def __norm_value_in_range(v, r):
        if r[1] == 0:
            return np.nan
        return (v - r[0]) / (r[1] - r[0])

    def __update_score_by_nOT(self, F):
        nOT = self.data["nOT"].values
        if nOT <= F[0]:
            self.data["score"] = 0
            return None
        if nOT > F[1]:
            self.data["score"] = np.inf
            return None
        return self.__norm_value_in_range(nOT, F)

    def __update_score_by_dG_Tm(self, Gs):
        ss_dG = self.data["ss_dG"].values
        tm_dG = self.data["tm_dG"].values
        if ss_dG >= tm_dG * min(Gs):
            self.data["score"] = 0
            return None
        if ss_dG < tm_dG * max(Gs):
            self.data["score"] = np.inf
            return None
        return self.__norm_value_in_range(ss_dG, [tm_dG * f for f in Gs])

    def __update_score_by_dG_Gs(self, Gs):
        ss_dG = self.data["ss_dG"].values
        if ss_dG >= Gs[0]:
            self.data["score"] = 0
            return None
        if ss_dG < Gs[1]:
            self.data["score"] = np.inf
            return None
        return self.__norm_value_in_range(ss_dG, Gs)

    def add_score(self, F, Gs):
        norm_nOT = self.__update_score_by_nOT(F)
        if norm_nOT is None:
            return
        if all(x >= 0 for x in Gs):
            norm_ss_dG = self.__update_score_by_dG_Tm(Gs)
        else:
            norm_ss_dG = self.__update_score_by_dG_Gs(Gs)
        if norm_ss_dG is None:
            return
        self.data["score"] = np.mean([norm_nOT, norm_ss_dG])


class OligoGroup:
    """Allows to select oligos from a group based on a "focus" window of
    interest. The window can be expanded to the closest oligo or to retain at
    least a given number of oligos.
    """

    _focus_window = None  # Left-close, right-open
    _oligos_in_focus_window = None
    _oligos_passing_score_filter = None

    def __init__(self, oligos, logger=logging.getLogger()):
        super(OligoGroup, self).__init__()
        self.logger = logger
        self._data = pd.concat(
            [o.to_dataframe() for o in oligos if np.isfinite(o["score"])],
            ignore_index=True,
        )
        self._data = self._data.loc[self._data["score"] <= 1, :]
        self._oligos_passing_score_filter = self._data.shape[0]

    @property
    def data(self):
        return self._data.copy()

    @property
    def focus_window(self):
        return self._focus_window

    @focus_window.setter
    def focus_window(self, focus_window):
        ass.ert_type(focus_window, tuple, "focus window")
        if len(focus_window) != 2:
            raise AssertionError
        if focus_window[1] <= focus_window[0]:
            raise AssertionError
        self._focus_window = focus_window

    @property
    def focus_window_repr(self):
        return f"[{self.focus_window[0]}:{self.focus_window[1]})"

    @property
    def focus_window_size(self):
        return self.focus_window[1] - self.focus_window[0]

    @property
    def oligos_in_focus_window(self):
        return self._oligos_in_focus_window

    @property
    def oligos_passing_score_filter(self):
        return self._oligos_passing_score_filter

    @property
    def usable_oligos(self):
        if isinstance(self.oligos_in_focus_window, type(None)):
            return self.oligos_passing_score_filter
        return np.logical_and(
            self.oligos_in_focus_window, self.oligos_passing_score_filter
        )

    def get_focused_oligos(self, onlyUsable=False):
        if onlyUsable:
            return self._data.loc[self.usable_oligos]
        return self._data.loc[self.oligos_in_focus_window, :]

    def get_n_focused_oligos(self, onlyUsable=False):
        if self.usable_oligos.sum() == 0:
            return 0
        return self.get_focused_oligos(onlyUsable).shape[0]

    def focus_all(self):
        # Use all oligos to define a focus region
        self.set_focus_window(
            self._data.loc[:, "start"].min(), self._data.loc[:, "start"].max() + 1
        )

    def set_focus_window(self, start, end, verbose=True):
        # Set a sub-window of interest to focus on
        self.focus_window = (start, end)

        start_condition = self._data["start"].values >= self.focus_window[0]
        end_condition = self._data["end"].values < self.focus_window[1]
        self._oligos_in_focus_window = np.logical_and(start_condition, end_condition)

        if verbose:
            nOligos = self.get_n_focused_oligos()
            nOligosUsable = self.get_n_focused_oligos(True)
            self.logger.info(
                " ".join(
                    [
                        f"Set focus region to {self.focus_window_repr}",
                        f"({nOligos} oligos, {nOligosUsable} usable)",
                    ]
                )
            )

    def expand_focus_to_n_oligos(self, n, verbose=True):
        # Expand the sub-window of interest to retrieve at least n oligos
        if isinstance(self.focus_window, type(None)):
            raise AssertionError

        if n <= self.get_n_focused_oligos():
            return

        for _ in range(n - self.get_n_focused_oligos()):
            if not self.expand_focus_to_closest():
                return

        if verbose:
            nOligos = self.get_n_focused_oligos()
            nOligosUsable = self.get_n_focused_oligos(True)
            self.logger.info(
                " ".join(
                    [
                        f"Expanded focus region to {self.focus_window}",
                        f" ({nOligos} oligos, {nOligosUsable} usable)",
                    ]
                )
            )

    def expand_focus_by_step(self, step, verbose=True):
        # Expand the current focus window of a given step (in nt)
        if step >= 0:
            raise AssertionError

        if (
            self.focus_window[0] <= self._data["start"].min()
            and self.focus_window[1] >= self._data["end"].max()
        ):
            self.logger.warning(
                " ".join(
                    [
                        "Cannot expand the focus region any further ",
                        "(all oligos already included)",
                    ]
                )
            )
            return False

        new_focus_start, new_focus_end = self.focus_window
        new_focus_start -= step / 2
        new_focus_end += step / 2
        new_focus_start = np.max([new_focus_start, self._data["start"].min()])
        new_focus_end = np.min([new_focus_end, self._data["end"].max()])

        self.set_focus_window(new_focus_start, new_focus_end, verbose)
        return True

    def __get_focus_extremes(self):
        earl = self._data["start"].values < self.focus_window[0]
        max_start = self._data.loc[earl, "start"].max()
        d_earl = self.focus_window[0] - max_start if earl.sum() != 0 else np.inf
        late = self._data["end"].values >= self.focus_window[1]
        min_end = self._data.loc[late, "end"].min()
        d_late = min_end - self.focus_window[1] if late.sum() != 0 else np.inf
        return (max_start, min_end, d_earl, d_late)

    def __set_focus_windows_for_inifinite_extreme(self):
        max_start, min_end, d_earl, d_late = self.__get_focus_extremes()
        if np.isinf(d_late):
            self.set_focus_window(max_start, self.focus_window[1], False)
        if np.isinf(d_earl):
            self.set_focus_window(self.focus_window[0], min_end + 1, False)

    def expand_focus_to_closest(self):
        # Expand the sub-window of interest to add the closest oligo
        # Return False if not possible (e.g., all oligos already included)
        if self.get_n_focused_oligos() == self._data.shape[0]:
            self.logger.warning(
                " ".join(
                    [
                        "Cannot expand the focus region any further",
                        "(all oligos already included)",
                    ]
                )
            )
            return False

        max_start, min_end, d_earl, d_late = self.__get_focus_extremes()

        if np.isinf(d_late) and np.isinf(d_earl):
            return False

        if np.isinf(d_late) or np.isinf(d_earl):
            self.__set_focus_windows_for_inifinite_extreme()
            return True

        if d_earl <= d_late:
            self.set_focus_window(max_start, self.focus_window[1], False)
        else:
            self.set_focus_window(self.focus_window[0], min_end + 1, False)

        return True

    def apply_threshold(self, threshold):
        # Unfocuses oligos with score higher than the threshold
        if threshold > 1 or threshold < 0:
            raise AssertionError
        self._oligos_passing_score_filter = self._data["score"] <= threshold

    def reset_threshold(self):
        self.apply_threshold(1)

    def discard_focused_oligos_safeDist(self, safeDist):
        # Discard focused oligos that are not within a safe distance from the
        # CFR borders
        start = self.data.loc[self.oligos_in_focus_window, "start"].min()
        end = self.data.loc[self.oligos_in_focus_window, "end"].max() + 1
        self.__discard_oligos_in_range(start + safeDist, end - safeDist)

    def __check_oligos_to_discard_safeN(self, safeN, start, end, D):
        oData = self.get_focused_oligos()
        c = 1
        while c <= safeN:
            passing_oData = oData.loc[oData["start"] > start, :]
            if passing_oData.shape[0] == 0:
                self.logger.info("Not enough oligos, skipped discard step.")
                return False
            start = passing_oData["start"].values[0]
            start += len(passing_oData["sequence"].values[0]) + D
            c += 1

        c = 1
        while c <= safeN:
            passing_oData = oData.loc[oData["end"] <= end]
            if passing_oData.shape[-1] == 0:
                self.logger.info("Not enough oligos, skipped discard step.")
                return False
            end = passing_oData["end"].values[-1]
            end -= len(passing_oData["sequence"].values[-1]) - D
            c += 1

        return True

    def discard_focused_oligos_safeN(self, safeN, D):
        # Discard focused oligos that are neither the first nor the last safeN
        start = self.data.loc[self.oligos_in_focus_window, "start"].values[0]
        start += (
            len(self.data.loc[self.oligos_in_focus_window, "sequence"].values[0]) + D
        )
        end = self.data.loc[self.oligos_in_focus_window, "end"].values[-1]
        end += (
            len(self.data.loc[self.oligos_in_focus_window, "sequence"].values[-1]) + D
        )

        if not self.__check_oligos_to_discard_safeN(safeN, start, end, D):
            return

        if not self.__discard_oligos_in_range(start, end):
            self.logger.info(f"No oligos to discard in range [{start}:{end}).")

    def __discard_oligos_in_range(self, start, end):
        if start >= end:
            return False
        start_condition = self.data["end"] < start
        end_condition = self.data["start"] >= end
        keep_condition = np.logical_or(start_condition, end_condition)
        nDiscarded = self.oligos_in_focus_window.sum() - keep_condition.sum()
        self.__apply_keep_condition(keep_condition)
        self.logger.info(
            f"Discarded {nDiscarded}" + f" oligos from the [{start}:{end}) range."
        )
        return True

    def __apply_keep_condition(self, keep_condition):
        self._data = self._data.loc[keep_condition, :]
        self._oligos_in_focus_window = self._oligos_in_focus_window[keep_condition]
        self._oligos_passing_score_filter = self._oligos_passing_score_filter[
            keep_condition
        ]


def select_by_GC(
    oligos_list: List[Tuple[str, str]],
    kmer_size: int,
    gc_range: GCRange = GCRange(0.35, 0.85),
) -> List[Tuple[str, str]]:
    return [
        (header, sequence)
        for header, sequence in tqdm(oligos_list, desc="GC check", leave=False)
        if (sequence.count("C") + sequence.count("G")) / kmer_size >= gc_range.low
        and (sequence.count("C") + sequence.count("G")) / kmer_size <= gc_range.high
    ]



# Expose the following functions to the main package
from .cycling_query import cycling_query
from .escafish_score import escafish_score
from .exclude_region import exclude_region
from .generate_exclude import generate_exclude
from .get_oligos import get_oligos
from .HUSH_feedback import HUSH_feedback
from .reform_hush_combined import reform_hush_combined
from .reform_hush_consec import reform_hush_consec
from .reform_hush import reform_hush
from .select_probe import select_probe
from .split_fasta import split_fasta
from .summarize_probes_cumul import summarize_probes_cumul
from .summarize_probes_final import summarize_probes_final
from .summarize_probes import summarize_probes


__all__ = ["cycling_query",
            "escafish_score",
            "exclude_region",
            "generate_exclude",
            "get_oligos",
            "HUSH_feedback",
            "reform_hush_combined",
            "reform_hush_consec",
            "reform_hush",
            "select_probe",
            "split_fasta",
            "summarize_probes_cumul",
            "summarize_probes_final",
            "summarize_probes"]


"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

import argparse
from ifpd2 import __version__
import sys

# import multiprocessing as mp


def add_version_option(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    parser.add_argument(
        "--version", action="version", version=f"{sys.argv[0]} {__version__}"
    )
    return parser



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

import copy
from ifpd2 import const
from ifpd2.io import get_dtype_length
import numpy as np  # type: ignore
import pandas as pd  # type: ignore
from rich.progress import Progress, TaskID  # type: ignore
from typing import Dict, List, Set, Tuple


class ChromosomeIndex:
    """ChromosomeIndex"""

    _bin_size: int
    _index: Dict[int, Tuple[float, float]]

    def __init__(self, bin_size: int):
        super(ChromosomeIndex, self).__init__()
        if bin_size < 1:
            raise AssertionError
        self._bin_size = bin_size

    def __init_index(self, chrom_db: pd.DataFrame) -> None:
        """Initialize chromosome index

        Creates an chromosome index with empty bins.

        Arguments:
            chrom_db {pd.DataFrame} -- chromosome database
        """
        self._index = {}
        chrom_size_nt = chrom_db["end"].values.max()
        for bin_id in range((chrom_size_nt // self._bin_size) + 1):
            self._index[bin_id] = (np.inf, 0)

    @property
    def bin_size(self) -> int:
        return self._bin_size

    @property
    def data(self) -> Dict[int, Tuple[float, float]]:
        return copy.copy(self._index)

    def __populate_bins(
        self,
        chrom_db: pd.DataFrame,
        record_byte_size: int,
        track: Tuple[Progress, TaskID],
    ) -> None:
        """Populate bins.

        Add extremities (byte positions) of each bin.

        Arguments:
            chrom_db {pd.DataFrame} -- chromosome database
            record_byte_size {int} -- size of a record in bytes (based on column dtype)
            track {Tuple[Progress, TaskID]} -- progress bar details
        """
        current_position = -1
        for i in range(chrom_db.shape[0]):
            track[0].update(track[1], advance=1)
            position_in_nt = chrom_db["start"].values[i]
            if position_in_nt <= current_position:
                raise AssertionError
            current_position = position_in_nt

            position_in_bytes = record_byte_size * i
            binned_to = position_in_nt // self._bin_size

            bin_start, bin_end = list(self._index[binned_to])
            bin_start = min(bin_start, position_in_bytes)
            bin_end = max(bin_end, position_in_bytes)
            self._index[binned_to] = (bin_start, bin_end)

    def __fill_empty_bins(self) -> None:
        """Populate empty bins of the index."""
        if not np.isfinite(self._index[0][0]):
            self._index[0] = (0, 0)
        for bin_id, (start, end) in self._index.items():
            if not np.isfinite(start):
                self._index[bin_id] = (
                    self._index[bin_id - 1][1],
                    self._index[bin_id - 1][1],
                )

    def build(
        self,
        chrom_db: pd.DataFrame,
        record_byte_size: int,
        track: Tuple[Progress, TaskID],
    ) -> None:
        """Build chromosome index.

        [description]

        Arguments:
            chrom_db {pd.DataFrame} -- chromosome database
            record_byte_size {int} -- size of a record in bytes (based on column dtype)
            track {Tuple[Progress, TaskID]} -- progress bar details
        """
        for colname in ("chromosome", "start", "end"):
            if colname not in chrom_db.columns:
                raise AssertionError(f"missing '{colname}' column")
        chromosome_set: Set[bytes] = set(chrom_db["chromosome"].values)
        if len(chromosome_set) != 1:
            raise AssertionError

        self.__init_index(chrom_db)
        self.__populate_bins(chrom_db, record_byte_size, track)
        self.__fill_empty_bins()

    def __getitem__(self, position_in_nt: int) -> int:
        if self._index is None:
            raise AssertionError
        binned_to = position_in_nt // self._bin_size
        if binned_to not in self._index:
            return -1

        position_in_bytes = int(self._index[binned_to][0])
        if not np.isfinite(position_in_bytes):
            return -1
        return position_in_bytes

    def __eq__(self, other) -> bool:
        condition = self.bin_size == other.bin_size
        condition = condition and self.data == other.data
        return condition


class ChromosomeData:
    """Contains information on a chromosome"""

    _name: bytes
    _size_nt: int
    _size_bytes: int
    _recordno: int
    _index: ChromosomeIndex
    _record_byte_size: int

    def __init__(
        self,
        chromosome_db: pd.DataFrame,
        dtype: Dict[str, str],
        index_bin_size: int,
        progress: Progress,
    ):
        super(ChromosomeData, self).__init__()

        if "chromosome" not in chromosome_db.columns:
            raise AssertionError
        selected_chrom = chromosome_db["chromosome"][0]
        if len(set(chromosome_db["chromosome"].values)) != 1:
            raise AssertionError

        self._record_byte_size = get_dtype_length(dtype)
        if self._record_byte_size <= 0:
            raise AssertionError

        self._name = selected_chrom
        self._recordno = chromosome_db.shape[0]
        self._size_nt = chromosome_db["end"].values.max()
        self._size_bytes = chromosome_db.shape[0] * self._record_byte_size

        self._build_index(chromosome_db, index_bin_size, progress)

    @property
    def name(self) -> bytes:
        return self._name

    @property
    def record_byte_size(self) -> int:
        return self._record_byte_size

    @property
    def size_nt(self) -> int:
        return self._size_nt

    @property
    def size_bytes(self) -> int:
        return self._size_bytes

    @property
    def recordno(self) -> int:
        return self._recordno

    @property
    def index(self) -> ChromosomeIndex:
        return copy.copy(self._index)

    def _build_index(
        self, chromosome_db: pd.DataFrame, index_bin_size: int, progress: Progress
    ) -> None:
        """Build a chromosome index.

        Build bin index based on chromosome length and bin size.

        Arguments:
            chromosome_db {pd.DataFrame} -- chromosome database
            index_bin_size {int} -- index bin size
            progress {Progress} -- progress instance for progress bar with rich
        """
        if index_bin_size <= 0:
            raise AssertionError
        indexing_track = progress.add_task(
            f"indexing {self._name.decode()}.bin",
            total=chromosome_db.shape[0],
            transient=True,
        )
        self._index = ChromosomeIndex(index_bin_size)
        self._index.build(
            chromosome_db, self._record_byte_size, (progress, indexing_track)
        )
        progress.remove_task(indexing_track)

    def __eq__(self, other) -> bool:
        condition = self.record_byte_size == other.record_byte_size
        condition = condition and self.size_nt == other.size_nt
        condition = condition and self.size_bytes == other.size_bytes
        condition = condition and self.recordno == other.recordno
        condition = condition and self.index == other.index
        return condition


class ChromosomeDict:
    """Wraps ChromosomeData instances into a dictionary
    and allow easier access to their attributes.
    """

    _index_bin_size: int
    _data: Dict[bytes, ChromosomeData]

    def __init__(self, index_bin_size: int = const.DEFAULT_DATABASE_INDEX_BIN_SIZE):
        super(ChromosomeDict, self).__init__()
        self._data = {}
        if index_bin_size <= 0:
            raise AssertionError
        self._index_bin_size = index_bin_size

    def __len__(self) -> int:
        return len(self._data)

    def keys(self) -> List[bytes]:
        return list(self._data.keys())

    def items(self) -> List[Tuple[bytes, ChromosomeData]]:
        return list(self._data.items())

    @property
    def index_bin_size(self) -> int:
        return self._index_bin_size

    @property
    def sizes_nt(self) -> Dict[bytes, int]:
        return {name: data.size_nt for name, data in self._data.items()}

    @property
    def sizes_bytes(self) -> Dict[bytes, int]:
        return {name: data.size_bytes for name, data in self._data.items()}

    @property
    def recordnos(self) -> Dict[bytes, int]:
        return {name: data.recordno for name, data in self._data.items()}

    def get_chromosome(self, chromosome: bytes) -> ChromosomeData:
        return copy.copy(self._data[chromosome])

    def add_chromosome(
        self, chromosome_db: pd.DataFrame, dtype: Dict[str, str], progress: Progress
    ) -> None:
        self._data[chromosome_db["chromosome"][0]] = ChromosomeData(
            chromosome_db, dtype, self._index_bin_size, progress
        )

    def __eq__(self, other) -> bool:
        condition = len(self) == len(other)
        condition = condition and self.items() == other.items()
        condition = condition and self.index_bin_size == other.index_bin_size
        return condition



#!/usr/bin/python3


# TO DO: 
# Switch to click arguments


from Bio.SeqIO.FastaIO import SimpleFastaParser
from ..ifpd2_new.ifpd2.scripts.extract_kmers import main as extract
import pandas as pd
import os
import sys
from tqdm import tqdm 
from tabulate import tabulate

# Retrieve complete sequences in ROI
def get_oligos(gcfilter:int = 1, extfolder:os.PathLike = './data/')->None:
    #syntax: ./get_oligos.py DNA/RNA gcfilter extfolder

    if(len(sys.argv) == 1):
        # no probe type was specified
        type = 'DNA'
    elif(len(sys.argv) == 2):
        type = sys.argv[1]
    elif(len(sys.argv) == 3):
        type = sys.argv[1]
        gcfilter = sys.argv[2]
    elif(len(sys.argv) == 4):
        type = sys.argv[1]
        gcfilter = sys.argv[2]
        extfolder = sys.argv[3]

    else:
         print(f'Incorrect number of arguments. Exiting...')
         exit(-1)


    if type=='DNA':
        roifile = extfolder+'rois/all_regions.tsv'
        ref = extfolder+'ref/'
        outseq = extfolder+'regions/'

        try:
            os.mkdir(outseq)
        except FileExistsError:
            print("Saving to existing 'regions' directory.")
            
        f = open(roifile)
        rd = pd.read_csv(f,sep="\t",header=0)

        for k in tqdm(range(len(rd)),desc='Retrieving sequences for all ROIs'):
            if not pd.isnull(rd.ref[k]):
                chrnr = rd.chrom[k][3:]
                reffile = ref+str(rd.at[k,'ref'])+'.chromosome.'+chrnr+'.fa'

                with open(reffile) as handle:
                    for values in SimpleFastaParser(handle):
                        fullseq = values[1]
                        seq = fullseq[rd.Window_start[k]-1:rd.Window_end[k]]         # shift index by 1 to match ref genome

                        # export sequences
                        out = open(outseq+'roi_'+str(rd.window_id[k])+'.fa','w')
                        out.write('>ROI_'+str(rd.window_id[k])+' pos='+rd.chrom[k]+':'+str(rd.Window_start[k])+'-'+str(rd.Window_end[k])+'\n'+seq)
                        out.close()       
                        
                        
        # Divide into k-mers
        outcan = extfolder+'candidates/'

        try:
            os.mkdir(outcan)
        except FileExistsError:
            print("Saving to existing 'candidates' directory.")
            
        for k in range(len(rd)):
            fullseq = outseq+'roi_'+str(rd.window_id[k])+'.fa'
            if not os.path.isfile(fullseq):
                print('The FASTA sequence for ROI '+str(rd.window_id[k])+' is missing.')
                continue
            extract(fullseq,outcan,rd.length[k],gcfilter)

    elif type == 'RNA':
        roifile = extfolder+'rois/all_regions.tsv'
        ref = extfolder+'ref/'
        outseq = extfolder+'regions/'

        # ASSUME THAT THE TRANSCRIPT SEQUENCES HAVE ALREADY BEEN IDENTIFIED
        # implement direct transcript retrieval? 

        f = open(roifile)
        rd = pd.read_csv(f,sep="\t",header=0)              
                        
        # Divide into k-mers
        outcan = extfolder+'candidates/'

        try:
            os.mkdir(outcan)
        except FileExistsError:
            print("Saving to existing 'candidates' directory.")
            
        for k in range(len(rd)):
            fullseq = outseq+'roi_'+str(rd.window_id[k])+'.fa'
            if not os.path.isfile(fullseq):
                print('The FASTA sequence for ROI '+str(rd.window_id[k])+' is missing.')
                continue
            extract(fullseq,outcan,rd.length[k],gcfilter)

    return   

if __name__ == '__main__':
    get_oligos()


"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

import configparser as cp
import copy
import logging
import multiprocessing as mp
import numpy as np  # type: ignore
import os
import pandas as pd  # type: ignore
from pathlib import Path
import shutil
from tqdm import tqdm  # type: ignore
from typing import Dict, List, Optional, Union

from ifpd2 import asserts as ass
from ifpd2.logging import add_log_file_handler
from ifpd2.database import DataBase
from ifpd2.oligo import OligoGroup
from ifpd2.probe import OligoProbeBuilder
from ifpd2.walker2 import ChromosomeWalker


class GenomicWindowSet:
    """Genomic window manager."""

    _window_sets = None

    C = "chr18"  # Chromosome
    S = int(3000000)  # Region start coordinate (included)
    E = int(3500000)  # Region end coordinate (excluded)

    X = 20  # Number of probes to design
    Ws: Optional[int] = None  # Window size (used when X is not provided)
    Wh: Optional[float] = 0.1  # Window shift (as a percentage of the window size)

    Rs: Union[int, float] = int(8000)  # Region focus size, either in nt (Rs > 1)
    #  or fraction of Ws (0<Rs<=1)
    #  When provided in nt, it is applied only if Rs<Ws
    Rt: Union[int, float] = int(1000)  # Region focus step, either in nt (Rt > 1)
    #  or fraction of Rs (0<Rt<=1),
    #  for focus region expansion
    _growing = False

    @property
    def window_sets(self):
        return self._window_sets

    @window_sets.setter
    def window_sets(self, window_sets):
        self._window_sets = pd.DataFrame(np.vstack(window_sets))
        self._window_sets.columns = [
            "start",
            "mid",
            "end",
            "cfr_start",
            "cfr_end",
            "w",
            "s",
        ]
        self._window_sets.sort_values("end", inplace=True)
        self._window_sets.reset_index(inplace=True, drop=True)

    @property
    def wid(self):
        return self._w

    @property
    def current_window(self):
        return self.window_sets.iloc[self.wid, :]

    @property
    def reached_last_window(self):
        return self._reached_last_window

    @property
    def growing(self):
        return self._growing

    def _assert(self):
        ass.ert_type(self.C, str, "C")
        ass.ert_type(self.S, int, "S")
        ass.ert_nonNeg(self.S, "S", True)
        ass.ert_type(self.E, int, "E")
        ass.ert_nonNeg(self.E, "E", True)
        if self.S > self.E:
            raise AssertionError

        ass.ert_multiTypes(self.X, [int, type(None)], "X")
        ass.ert_multiTypes(self.Ws, [int, type(None)], "Ws")
        if isinstance(self.X, int):
            ass.ert_type(self.Ws, type(None), "Ws")
            ass.ert_nonNeg(self.X, "X")
        else:
            ass.ert_type(self.Ws, int, "Ws")

        ass.ert_multiTypes(self.Ws, [int, type(None)], "Ws")
        if isinstance(self.Ws, int):
            ass.ert_type(self.X, type(None), "X")
            ass.ert_nonNeg(self.Ws, "Ws")
        else:
            ass.ert_type(self.X, int, "X")

        ass.ert_type(self.Wh, float, "Wh")
        ass.ert_inInterv(self.Wh, 0, 1, "Wh")

        ass.ert_multiTypes(self.Rs, [int, float], "Rs")
        if self.Rs > 1:
            self.Rs = int(self.Rs)
        else:
            ass.ert_inInterv(self.Rs, 0, 1, "Rs")

        ass.ert_multiTypes(self.Rt, [int, float], "Rt")
        if self.Rt > 1:
            self.Rt = int(self.Rt)
        else:
            ass.ert_inInterv(self.Rt, 0, 1, "Rt")

    def _init_windows(self):
        # Build windows and central focus regions (CFR)
        if not os.path.isdir(os.path.join(self.out_path, "window_sets")):
            os.mkdir(os.path.join(self.out_path, "window_sets"))

        if self.S == self.E and isinstance(self.Ws, type(None)):
            raise AssertionError(
                " ".join(
                    [
                        "During full-chromosome search, provide a window size.",
                        "I.e., it is not possible to design X probes.",
                    ]
                )
            )

        if isinstance(self.X, int):
            self.Ws = (
                self.E - self.S
                if self.X == 1
                else np.floor((self.E - self.S) / (self.X + 1)).astype("i")
            )

        self._w = 0  # Window ID
        self._reached_last_window = False

        if self.S == self.E:
            self._growing = True
            self.window_sets = self.__mk_first_window()
        else:
            self.window_sets = self.__mk_all_window_sets()

    def __update_focus(self):
        if isinstance(self.Rs, float):
            self.Rs = int(self.Rs * self.Ws)
        if isinstance(self.Rt, float):
            self.Rt = int(self.Rs * self.Rt)

    def __mk_all_window_sets(self):
        # Prepare all window sets in a region of interest
        window_starts = np.floor(np.arange(self.S, self.E, self.Ws)).astype("i")
        if (self.E - self.S) % self.Ws != 0 or len(window_starts) != 1:
            window_starts = window_starts[:-1]

        window_mids = (window_starts + self.Ws / 2).reshape((window_starts.shape[0], 1))
        window_borders = np.transpose(
            np.vstack([window_starts, window_starts + self.Ws])
        )

        nWindows = window_borders.shape[0]

        self.__update_focus()
        if self.Rs < self.Ws:
            central_regions = np.hstack(
                [
                    np.floor(window_mids - self.Rs / 2),
                    np.floor(window_mids + self.Rs / 2),
                ]
            )
        else:
            nans = np.array([np.nan for x in range(window_mids.shape[0])])
            central_regions = np.vstack([nans, nans]).transpose()

        window_sets = []
        for i in range(len(np.arange(0, 1, self.Wh))):
            winID = np.array(range(nWindows)).reshape((nWindows, 1))
            setID = np.repeat(i, nWindows).reshape((nWindows, 1))
            window_sets.append(
                np.hstack(
                    [
                        window_borders + i * self.Wh * self.Ws,
                        window_mids + i * self.Wh * self.Ws,
                        central_regions + i * self.Wh * self.Ws,
                        winID,
                        setID,
                    ]
                )[:, (0, 2, 1, 3, 4, 5, 6)]
            )

        return window_sets

    def __mk_first_window(self):
        # Initialize only the first window
        mid = self.S + self.Ws / 2
        if self.Rs < self.Ws:
            cstart, cend = (np.floor(mid - self.Rs / 2), np.floor(mid + self.Rs / 2))
        else:
            cstart, cend = (np.nan, np.nan)
        return [[self.S, mid, self.S + self.Ws, cstart, cend, 0, 0]]

    def _add_window(self):
        # Add a window, assigning it to the first set with no overlaps
        new_window = self.window_sets.iloc[-1, :].copy()
        new_window.iloc[:5] = new_window.iloc[:5] + self.Wh * self.Ws

        gData = np.array(
            [(n, g["end"].max()) for n, g in self.window_sets.groupby("s")]
        )
        non_overlapping_group_ids = np.where(gData[:, 1] <= new_window[0])[0]

        new_window_id = 0
        new_set_id = self.window_sets["s"].max() + 1
        if len(non_overlapping_group_ids) != 0:
            new_set_id = gData[non_overlapping_group_ids].min()
            new_window_id = (
                self.window_sets[self.window_sets["s"] == new_set_id]["w"].max() + 1
            )

        new_window.iloc[5] = new_window_id
        new_window.iloc[6] = new_set_id
        self.window_sets = [self.window_sets.values, new_window.values.reshape((1, 7))]

    def go_to_next_window(self):
        if self._growing:
            self._add_window()
            self._w += 1
        else:
            if self.wid < self.window_sets.shape[0] - 1:
                self._w += 1
            if self.wid >= self.window_sets.shape[0] - 1:
                self._reached_last_window = True

    def export_window_set(self):
        self.window_sets.loc[
            self.window_sets["s"] == self.current_window["s"], :
        ].to_csv(os.path.join(self.window_set_path, "windows.tsv"), "\t")


class Walker(GenomicWindowSet):
    """Walker walks through the oligos stored in an ifpd2 database,
    assigns them to windows based on user-defined parameters, and then builds
    probe candidates.
    """

    out_path = "."
    reuse = False
    _threads = 1
    __db: DataBase

    __current_oligos: List = []
    __walk_results: Dict = {}

    def __init__(self, db_path, logger=logging.getLogger()):
        GenomicWindowSet.__init__(self)
        self.__db = DataBase(db_path)
        self.logger = logger

    @property
    def threads(self):
        return self._threads

    @threads.setter
    def threads(self, threads):
        ass.ert_type(threads, int, threads)
        threads = max(1, min(threads, mp.cpu_count()))
        self._threads = threads

    @property
    def current_oligos(self):
        return self.__current_oligos

    def remove_oligos_starting_before(self, pos):
        self.__current_oligos = [o for o in self.current_oligos if o["start"] >= pos]

    @property
    def window_set_path(self):
        return os.path.join(
            self.out_path, "window_sets", f"set_{int(self.current_window['s'])}"
        )

    @property
    def window_path(self):
        return os.path.join(
            self.window_set_path, f"window_{int(self.current_window['w'])}"
        )

    @property
    def window_tag(self):
        window = self.current_window
        return f"{int(window['s'])}.{int(window['w'])}"

    @property
    def window_range(self):
        window = self.current_window
        return f"[{int(window['start'])}:{int(window['end'])}]"

    @property
    def walk_results(self):
        return self.__walk_results

    @property
    def config(self):
        config = cp.ConfigParser()

        config["AIM"] = {"Region": f"{self.C}:{self.S}-{self.E}"}
        if not isinstance(self.X, type(None)):
            config["AIM"]["Probe(s) number"] = str(self.X)

        config["WINDOWS"] = {}
        if not isinstance(self.Ws, type(None)):
            config["WINDOWS"]["Window size"] = str(self.Ws)
        config["WINDOWS"].update(
            {
                "Window step": str(self.Wh),
                "Focus region size": str(self.Rs),
                "Focus region step": str(self.Rt),
            }
        )
        return config

    def _assert(self):
        GenomicWindowSet._assert(self)
        if not os.path.isdir(self.out_path):
            raise AssertionError

    def start(self, *args, start_from_nt: int = 0, end_at_nt: int = -1, **kwargs):
        self._assert()
        self._init_windows()
        self.print_prologue()
        self.__start_walk(
            *args, start_from_nt=start_from_nt, end_at_nt=end_at_nt, **kwargs
        )

    def print_prologue(self):

        s = "* Walker *\n\n"
        s += f"Threads: {self.threads}\n"
        s += f"Database: '{self.__db.path}'\n"
        s += f"Region of interest: {self.C}:{self.S}-{self.E}\n"

        if not isinstance(self.X, type(None)):
            nWindows = int(self.window_sets["w"].max() + 1)
            s += f"Aim to scout {nWindows} windows.\n\n"

        s += f"Using a central focus region of {self.Rs} nt,"
        s += f" in windows of size {self.Ws} nt,\n"
        s += f"built with a shift of {self.Ws*self.Wh} nt ({self.Wh*100}%).\n"
        if not isinstance(self.X, type(None)):
            nSets = int(self.window_sets["s"].max() + 1)
            s += f"Thus, a total of {nSets} window sets will be explored.\n"

        self.logger.info(s)

    @property
    def walk_destination(self):
        if self.E == self.S:
            return np.inf
        return self.E

    def __finished_parsing(self, record, DBHpb, args, kwargs):
        # Return:
        #   None to stop parsing
        #   True to parse in current window
        #   Parsed output to append it

        if record["start"] < self.current_window["end"]:
            return (None, "continue")
        # DBHpb.clear()

        parsed_output = self.__process_window_async(
            self.current_oligos,
            self.current_window,
            self.fprocess,
            self.fpost,
            *args,
            opath=self.window_path,
            loggerName=self.logger.name,
            **kwargs,
        )

        if self.reached_last_window:
            self.logger.info("Reached last window")
            return (None, "break")

        self.go_to_next_window()
        self._preprocess_window()
        self._load_windows_until_next_to_do()
        if self.reached_last_window and self._window_done():
            self.logger.info("Reached last window and done")
            return (None, "break")

        if len(self.current_oligos) != 0:
            self.remove_oligos_starting_before(self.current_window["start"])

        return (parsed_output, "append")

    def __parse_line(self, record, DBHpb, args, kwargs):
        if record["start"] >= self.current_window["start"]:
            parsing_output, parsing_status = self.__finished_parsing(
                record, DBHpb, args, kwargs
            )
            if parsing_status != "continue":
                return (parsing_output, parsing_status)
            if record["end"] > self.walk_destination:  # End reached
                self.logger.info("Reached destination")
                return (None, "break")

            self.fparse(record, *args, **kwargs)

            if not np.isnan(record["score"]):
                self.current_oligos.append(record)

            self.rw += 1
        return (None, "continue")

    def __parse_database(self, DBHpb, args, kwargs):
        exec_results = []
        for record in DBHpb:
            parsing_output, parsing_status = self.__parse_line(
                record, DBHpb, args, kwargs
            )

            if parsing_status == "break":
                break
            if parsing_status == "append":
                exec_results.append(parsing_output)
            self.r += 1

        if len(self.current_oligos) != 0:
            exec_results.append(
                self.__process_window_async(
                    self.current_oligos,
                    self.current_window,
                    self.fprocess,
                    self.fpost,
                    *args,
                    opath=self.window_path,
                    loggerName=self.logger.name,
                    **kwargs,
                )
            )
        return exec_results

    def __end_walk(self, exec_results):
        for promise in exec_results:
            s, w, results = promise.get()
            if s in self.__walk_results.keys():
                self.__walk_results[s][w] = results
            else:
                self.__walk_results[s] = {w: results}

    def __start_walk(
        self, *args, start_from_nt: int = 0, end_at_nt: int = -1, **kwargs
    ):
        self.pool = mp.Pool(max(1, min(self.threads, mp.cpu_count())))
        self.logger.info(f"Prepared a pool of {self.threads} threads.")

        self._preprocess_window()
        self._load_windows_until_next_to_do()

        if self.reached_last_window and self._window_done():
            self.logger.info("All windows pre-processed. Skipped database walk.")
            return

        self.r = 0  # Record ID
        self.rw = 0  # Walk step counter

        DBHpb = tqdm(
            ChromosomeWalker(self.__db, self.C.encode()).buffer(
                start_from_nt, end_at_nt
            ),
            total=self.__db.chromosome_recordnos[self.C.encode()],
            desc="Parsing records",
            leave=False,
        )
        parsing_output = self.__parse_database(DBHpb, args, kwargs)
        self.logger.info(f"Parsed {self.rw}/{self.r} records.")

        self.__end_walk(parsing_output)
        DBHpb.close()
        self.pool.close()

    def _window_done(self):
        s = int(self.current_window["s"])
        w = int(self.current_window["w"])
        if s in self.walk_results.keys() and w in self.walk_results[s].keys():
            return isinstance(self.walk_results[s][w], list)
        return False

    def _load_windows_until_next_to_do(self):
        while self._window_done() and not self.reached_last_window:
            self.go_to_next_window()
            self._preprocess_window()

    def _preprocess_window_do(self):
        win_file_path = os.path.join(self.window_path, "window.tsv")
        win_done = os.path.isfile(os.path.join(self.window_path, ".done"))

        if os.path.isfile(win_file_path) and win_done:
            win = pd.read_csv(win_file_path, sep="\t", header=None, index_col=0)

            if (win.transpose().values[0][1:] == self.current_window.values).all():
                self.logger.info(
                    " ".join(
                        [
                            "Re-using previous results for window",
                            f"{self.window_tag} {self.window_range}",
                        ]
                    )
                )

                sid = int(self.current_window["s"])
                if sid not in self.walk_results.keys():
                    self.__walk_results[sid] = {}
                wid = int(self.current_window["w"])
                self.__walk_results[sid][wid] = self.fimport(self.window_path)
                return
        shutil.rmtree(self.window_path)
        os.mkdir(self.window_path)

    def _preprocess_window(self):
        # Preprocess current window
        # Triggers import if previously run and matching current window

        if not os.path.isdir(self.window_set_path):
            os.mkdir(self.window_set_path)
        self.export_window_set()

        if not os.path.isdir(self.window_path):
            os.mkdir(self.window_path)
        else:
            if not self.reuse:
                shutil.rmtree(self.window_path)
                os.mkdir(self.window_path)

            self._preprocess_window_do()

        self.current_window.to_csv(
            os.path.join(self.window_path, "window.tsv"), sep="\t", index=True
        )
        with open(os.path.join(self.window_path, "walker.config"), "w+") as CPH:
            self.config.write(CPH)

    @staticmethod
    def process_window_async(
        oligos,
        window,
        fprocess,
        fpost,
        *args,
        N=1,
        opath=None,
        loggerName=None,
        **kwargs,
    ):
        # Wrapper of process_window function, for parallelization
        window_tag = f"{int(window['s'])}.{int(window['w'])}"

        logger_name = f"ifpd2-window-{window_tag}"
        logger = logging.getLogger(logger_name)
        logger.propagate = False
        logger.setLevel(logging.DEBUG)
        logger.handlers = []
        add_log_file_handler("{0}/{1}.log".format(opath, "window"), logger.name)

        mainLogger = logging.getLogger(loggerName)
        mainLogger.info(f"Window {window_tag} sent to pool.")
        results = Walker.process_window(
            oligos,
            window,
            fprocess,
            fpost,
            *args,
            N=N,
            opath=opath,
            loggerName=logger.name,
            **kwargs,
        )
        mainLogger.info(f"Processor pool returned window {window_tag}.")
        return results

    def __process_window_async(self, *args, **kwargs):
        if self.pool is not None:
            return self.pool.apply_async(Walker.process_window_async, args, kwargs)
        return None

    @staticmethod
    def process_window(
        oligos,
        window,
        fprocess,
        fpost,
        *args,
        N=1,
        opath=None,
        loggerName=None,
        **kwargs,
    ):
        # Process oligos from window using fprocess. Then, post-process them
        # with fopost. Requires at least N oligos to proceed. If opath is
        # specified, a ".done" file is touched upon successful postprocessing.
        logger = logging.getLogger(loggerName)
        kwargs["loggerName"] = loggerName

        if len(oligos) >= N:
            oGroup = OligoGroup(oligos, logger)
            logger.info(
                " ".join(
                    [
                        f"Retrieved {oGroup.data.shape[0]} oligos for",
                        f"window {int(window['s'])}.{int(window['w'])}",
                        f"[{int(window['start'])}:{int(window['end'])}]",
                    ]
                )
            )
            results = fprocess(oGroup, window, *args, **kwargs)
        else:
            logger.warning(
                " ".join(
                    [
                        f"Window {int(window['s'])}.{int(window['w'])}",
                        "does not have enough oligos",
                        f"{len(oligos)}/{N}, skipped.",
                    ]
                )
            )
            results = []

        status, results = fpost(results, opath, *args, **kwargs)
        if status and opath is not None:
            Path(os.path.join(opath, ".done")).touch()

        return (int(window["s"]), int(window["w"]), results)

    @staticmethod
    def fparse(record, opb=None, *args, **kwargs):
        record.add_score(opb.F, opb.Gs)
        # if record["off_target_no"]<100:
        #     print((record._data, opb.F, opb.Gs))
        #     print(record.to_dataframe())

    @staticmethod
    def fprocess(oGroup, window, *args, **kwargs):
        opb = copy.copy(kwargs["opb"])
        if not isinstance(opb, OligoProbeBuilder):
            raise AssertionError
        logger = logging.getLogger(kwargs["loggerName"])
        probe_list = opb.start(oGroup, window, kwargs["cfr_step"], logger)
        reduced_probe_list = opb.reduce_probe_list(probe_list)
        logger.info(
            f"Reduced from {len(probe_list)} to " + f"{len(reduced_probe_list)} probes."
        )
        return reduced_probe_list

    @staticmethod
    def fimport(path, *args, **kwargs):
        return OligoProbeBuilder.import_probes(path)

    @staticmethod
    def fpost(results, opath, *args, **kwargs):
        if not isinstance(kwargs["opb"], OligoProbeBuilder):
            raise AssertionError
        logger = logging.getLogger(kwargs["loggerName"])
        if len(results) == 0:
            logger.critical(f"Built {len(results)} oligo probe candidates")
            logger.handlers = logger.handlers[:-1]
        else:
            logger.info(f"Built {len(results)} oligo probe candidates")
            logger.handlers = logger.handlers[:-1]
            OligoProbeBuilder.export_probes(results, opath)
        with open(os.path.join(opath, "builder.config"), "w+") as CPH:
            kwargs["opb"].config.write(CPH)
        return (True, results)



#!/usr/bin/python3

# Excludes a certain region from a reference chromosome.
# Used to run HUSH on particularly repetitive sequences
# check for matching oligos outside of the region of interest and its repeats

# No arguments
# Exclude regions according to bedfile in the "./data/exclude" folder, called excl_roi_#
# Fetches the chromosome in the "./data/ref/" folder and generates a genome ref file with that region masked

from Bio.SeqIO.FastaIO import SimpleFastaParser
from ..ifpd2_new.ifpd2.scripts.extract_kmers import main as extract
import pandas as pd
import os
import sys
from tqdm import tqdm
from tabulate import tabulate


def exclude_region(roifile:os.PathLike = './data/rois/all_regions.tsv',     # proceed all ROIs as provided in region list
                   ref: os.PathLike = './data/ref/',
                   excl: os.PathLike = './data/exclude/')->None:

    f = open(roifile)
    rd = pd.read_csv(f, sep="\t", header=0)

    for k in tqdm(range(len(rd)), desc='Processing all ROIs'):
        # fetch bed file with sequences to exclude
        bed = excl + 'excl_roi_' + str(rd.window_id[k]) + '.bed'

        if os.path.isfile(bed):   
            # load bed file corresponding to the current roi  
            bd = pd.read_csv(bed, delimiter='\t',header=0)

            # create a dictionary of all chromosomes
            dictseq = {}
            files = os.listdir(ref)
            for i in range(len(files)):
                with open(ref+files[i]) as handle:
                    if files[i][0:3] != "gen":
                        seq = ""
                        for values in SimpleFastaParser(handle):
                            seq = seq + values[1]  
                        dictseq[files[i]] = seq 
                    else:   # exclude genome ref files from dictionary
                        dictseq[files[i]] = ""    


            # edit chromosomes according to bed file
            for j in range(len(bd)):
                # in case the bed file refers to chromosomes not found in the ref folder
                if os.path.isfile(ref+bd.chrom[j]+".fa"):
                    # instead of cutting out the repeats, replace with NNNN (avoids index shifting issues)
                    current = dictseq[bd.chrom[j]+".fa"]
                    insert = "N"*(bd.chromEnd[j]-bd.chromStart[j]+1)
                    new = current[:bd.chromStart[j]] + insert + current[bd.chromEnd[j]+1:]      # indexing: T[a:b] a included b not
                    # replace in chromosome dictionary
                    dictseq[bd.chrom[j]+".fa"] = new
                else:
                    print("The chromosome "+bd.chrom[j]+".fa was not found in the ref folder (ROI "+str(rd.window_id[k])+"). Skipping.")    

            # concatenate chromosomes into new ref genome
            genome = ""
            for c in range(len(files)):
                genome = genome + dictseq[files[c]]

            # export new ref genome
            out = open(ref+'genome_roi_' + str(rd.window_id[k])+'.fa', 'w')
            out.write('>Genome_ROI_'+str(rd.window_id[k])+'\n'+genome)
            out.close()
    return

if __name__ == "__main__":
    exclude_region()



# Expose this functions to outside as a module
from .src import (
    cycling_query,
    escafish_score,
    exclude_region,
    generate_exclude,
    get_oligos,
    HUSH_feedback,
    reform_hush_combined,
    reform_hush_consec,
    reform_hush,
    select_probe,
    split_fasta,
    summarize_probes_cumul,
    summarize_probes_final,
    summarize_probes)

__all__ = ["cycling_query",
            "escafish_score",
            "exclude_region",
            "generate_exclude",
            "get_oligos",
            "HUSH_feedback",
            "reform_hush_combined",
            "reform_hush_consec",
            "reform_hush",
            "select_probe",
            "split_fasta",
            "summarize_probes_cumul",
            "summarize_probes_final",
            "summarize_probes"]


"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

import click  # type: ignore
from ifpd2 import const
from ifpd2.database import DataBase
from ifpd2.walker2 import ChromosomeWalker
from tqdm import tqdm  # type: ignore
from typing import List, Optional, Tuple


@click.command(
    name="dump",
    context_settings=const.CONTEXT_SETTINGS,
    help="""Dump whole INPUT database to the terminal, in tsv format.
    To dump a single feature, specify it with the --chrom option.
    To dump a region of a feature, also use the --start and --end options.""",
)
@click.argument("input_paths", metavar="INPUT", nargs=1, type=click.Path(exists=True))
@click.option("--chrom", type=click.STRING, help="Database chromosome to dump from.")
@click.option("--start", type=click.INT, help="Start location of the region to dump.")
@click.option("--end", type=click.INT, help="End location of the region to dump.")
def main(
    input_paths: str,
    chrom: Optional[str] = None,
    start: Optional[int] = None,
    end: Optional[int] = None,
) -> None:
    DB = DataBase(input_paths)
    chrom, start, end = check_region(DB, chrom, start, end)

    print("\t".join(const.database_columns))
    for chromosome in get_chromosome_list(DB, chrom):
        walker = ChromosomeWalker(DB, chromosome)
        for record in tqdm(
            walker.buffer(start, end),
            desc=f"dumping '{chromosome.decode()}'",
            total=DB.chromosome_recordnos[chromosome],
        ):
            if start > record["start"]:
                continue
            print(record.to_csv("\t"))


def check_region(
    DB: DataBase,
    chrom: Optional[str] = None,
    start: Optional[int] = None,
    end: Optional[int] = None,
) -> Tuple[Optional[str], int, int]:
    if chrom is None:
        if start is not None or end is not None:
            raise AssertionError(
                "cannot use --region-start or --region-end without --chrom"
            )
    elif start is not None:
        chrom_size = DB.chromosome_sizes_nt[chrom.encode()]
        if start >= chrom_size:
            raise AssertionError(f"{start} larger than chromosome size: {chrom_size}")
        if end is not None and start >= end:
            raise AssertionError("end location smaller than start")

    return (
        chrom,
        0 if start is None else start,
        -1 if end is None else end,
    )


def get_chromosome_list(DB: DataBase, chrom: Optional[str]) -> List[bytes]:
    chromosome_list = DB.chromosome_list
    if chrom is not None:
        if chrom.encode() not in chromosome_list:
            raise AssertionError(f"'{chrom}' not found")
        chromosome_list = [chrom.encode()]
    return chromosome_list



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

from ifpd2.scripts import autocomplete, ifpd2
from ifpd2.scripts import db, extract_kmers
from ifpd2.scripts import query, query2

import logging
from rich.logging import RichHandler  # type: ignore

logging.basicConfig(
    level=logging.INFO,
    format="%(message)s",
    handlers=[RichHandler(markup=True, rich_tracebacks=True)],
)

__all__ = [
    "autocomplete",
    "ifpd2",
    "db",
    "extract_kmers",
    "query",
    "query2",
]



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

from ifpd2.scripts.db import check, dump, info, make
from ifpd2.scripts.db import run, settings

__all__ = ["check", "dump", "info", "make", "run", "settings"]



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

import click  # type: ignore
from ifpd2.const import CONTEXT_SETTINGS
from ifpd2.scripts.db import check, dump, info, make


@click.group(
    name="db",
    context_settings=CONTEXT_SETTINGS,
    help="Tools to manage ifpd2 databases.",
)
def main() -> None:
    """Just a hook for the entry point. Silence is golden!"""


main.add_command(check.main)
main.add_command(dump.main)
main.add_command(info.main)
main.add_command(make.main)



#!/usr/bin/python3

from Bio.SeqIO.FastaIO import SimpleFastaParser
from ..ifpd2_new.ifpd2.scripts.extract_kmers import main as extract
import pandas as pd
import os
import sys
from tqdm import tqdm
from tabulate import tabulate


def split_fasta()->None:
    file = sys.argv[1]
    outfolder = sys.argv[2]

    with open(file) as handle:
        for values in SimpleFastaParser(handle):
            out = open(outfolder+values[0]+".fa", 'w')
            out.write('>'+values[0]+'\n'+values[1])
            out.close()

if __name__ == "__main__":
    split_fasta()


"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

from ifpd2.chromosome import ChromosomeData
import logging
from typing import List, Tuple


class GenomicRegion:
    """Details on genomic region with a central focus region."""

    __chrom: bytes
    __chromStart: int
    __chromEnd: int
    __focusStart: int
    __focusEnd: int
    __focusSize: int
    __focusStep: int

    def __init__(
        self,
        chrom_region: Tuple[bytes, int, int],
        focus_style: Tuple[float, float] = (1, 1),
    ):
        super(GenomicRegion, self).__init__()
        self.__init_region(*chrom_region)
        self.__init_focus(*focus_style)

    def __init_region(self, chrom: bytes, chromStart: int, chromEnd: int) -> None:
        """Initialize region coordinates.

        Arguments:
            chrom {bytes} -- chromosome name
            chromStart {int} -- chromosome start position (usually 0)
            chromEnd {int} -- chromosome end position
        """
        if len(chrom) == 0:
            raise AssertionError("chromosome cannot be empty")
        if chromStart < 0:
            raise AssertionError(
                f"start should be greater than or equal to 0: {chromStart}"
            )
        if chromEnd <= chromStart:
            raise AssertionError(
                f"end should be greater than start: {chromStart}-{chromEnd}"
            )
        self.__chrom = chrom
        self.__chromStart = chromStart
        self.__chromEnd = chromEnd

    def __init_focus_size(self, focus_style: float) -> None:
        """Initialize focus region size.

        Set the size of the central focus region. When focus style is lower than or
        equal to 1, it is considered to be a fraction of the genomic region size. When
        it is greater than 1, instead, it is considered to be the size in nt.

        Arguments:
            focus_style {float} -- focus region size in nt
                                   or as a fraction of the genomic region
        """
        if focus_style <= 0:
            raise AssertionError
        if focus_style > 1:
            if focus_style > self.__chromEnd - self.__chromStart:
                raise AssertionError
            self.__focusSize = int(focus_style)
        else:
            self.__focusSize = int((self.__chromEnd - self.__chromStart) * focus_style)

    def __init_focus_step(self, step_style: float) -> None:
        """Initialize step for focus region growth.

        Set the step at which the focus region grows. When step style is lower than or
        equal to 1, it is considered to be a fraction of the focus region size. When
        it is greater than 1, instead, it is considered to be in nt.

        Arguments:
            step_style {float} -- focus region growth step in nt
                                  or as a fraction of the focus region
        """
        if step_style <= 0:
            raise AssertionError
        if step_style > 1:
            self.__focusStep = int(step_style)
        else:
            self.__focusStep = int(self.__focusSize * step_style)

    def __init_focus(self, focus_style: float, focus_step_style: float) -> None:
        """Initialize focus region.

        More details on focus and focus step style in the __init_focus_size and
        __init_focus_step functions.

        Arguments:
            focus_style {float} -- focus region size in nt
                                   or as a fraction of the genomic region
            focus_step_style {float} -- focus region growth step in nt
                                        or as a fraction of the focus region
        """
        self.__init_focus_size(focus_style)
        self.__init_focus_step(focus_step_style)
        self.__focusStart = int(
            (self.__chromStart + self.__chromEnd) / 2 - self.__focusSize / 2
        )
        self.__focusEnd = int(
            (self.__chromStart + self.__chromEnd) / 2 + self.__focusSize / 2
        )

    @property
    def chromosome(self) -> bytes:
        return self.__chrom

    @property
    def start(self) -> int:
        return self.__chromStart

    @property
    def end(self) -> int:
        return self.__chromEnd

    @property
    def region(self) -> Tuple[int, int]:
        return (self.__chromStart, self.__chromEnd)

    @property
    def focus(self) -> Tuple[int, int]:
        return (self.__focusStart, self.__focusEnd)

    # @property
    # def focus_step(self) -> int:
    #     return self.__focusStep

    def can_increase_focus(self) -> bool:
        """Check if the focus region can grow more

        Returns:
            bool -- whether growth is possible
        """
        return self.focus != self.region

    def increase_focus(self) -> None:
        """Grow the focus region."""
        if self.can_increase_focus():
            logging.warning("cannot increase the focus region any further")
        self.__focusStart = max(
            self.__chromStart, self.__focusStart - self.__focusStep // 2
        )
        self.__focusEnd = min(self.__chromEnd, self.__focusEnd + self.__focusStep // 2)


class GenomicRegionBuilder:
    """docstring for GenomicRegionBuilder"""

    __chromosome: bytes
    __chromosome_size_nt: int
    __focus_style: Tuple[float, float]

    def __init__(
        self,
        chromosome_data: ChromosomeData,
        focus_style: float = 1,
        focus_step_style: float = 1,
    ):
        super(GenomicRegionBuilder, self).__init__()
        self.__chromosome = chromosome_data.name
        self.__chromosome_size_nt = chromosome_data.size_nt
        if focus_style <= 0:
            raise AssertionError
        if focus_step_style <= 0:
            raise AssertionError
        self.__focus_style = (focus_style, focus_step_style)

    def __build_overlapping(self, size: int, step: int) -> List[List[GenomicRegion]]:
        """Build lists of regions.

        Windows are generated based on size and step. Step should be smaller than size.
        Overlapping regions are placed in separate lists.

        Arguments:
            size {int} -- region size in nt
            step {int} -- region step in nt

        Returns:
            List[List[GenomicRegion]] -- generated region lists
        """
        if step >= size:
            raise AssertionError
        region_set_list: List[List[GenomicRegion]] = []
        for range_start in range(0, size, step):
            genomic_region_set: List[GenomicRegion] = []
            for start in range(range_start, self.__chromosome_size_nt, step):
                end = start + size
                if end < self.__chromosome_size_nt:
                    genomic_region_set.append(
                        GenomicRegion(
                            (self.__chromosome, start, end), self.__focus_style
                        )
                    )
            region_set_list.append(genomic_region_set)
        return region_set_list

    def __build_non_overlapping(
        self, size: int, step: int
    ) -> List[List[GenomicRegion]]:
        """Build a list of non-overlapping regions.

        Arguments:
            size {int} -- region size in nt
            step {int} -- region step in nt

        Returns:
            List[List[GenomicRegion]] -- generated region lists
        """
        if step != size:
            raise AssertionError
        genomic_region_set: List[GenomicRegion] = []
        for start in range(0, self.__chromosome_size_nt, step):
            end = start + size
            if end <= self.__chromosome_size_nt:
                genomic_region_set.append(
                    GenomicRegion((self.__chromosome, start, end), self.__focus_style)
                )
        return [genomic_region_set]

    def build_by_number(self, n: int) -> List[List[GenomicRegion]]:
        """Build a number of regions.

        Build a number of region instances covering the current chromosome.

        Arguments:
            n {int} -- number of regions

        Returns:
            List[List[GenomicRegion]] -- generated region lists
        """
        step: int = self.__chromosome_size_nt // n
        return self.build_by_size(step, step)

    def build_by_size(self, size: int, step_style: float) -> List[List[GenomicRegion]]:
        """Build regions by size and step (allows for overlap).

        Build region instances based on their size and step. Step is the distance
        between the start of a region and the start of the next region. Overlapping
        regions are placed in separate lists.

        Arguments:
            size {int} -- region size in nt
            step_style {float} -- region step in nt or as a fraction of the region size

        Returns:
            List[List[GenomicRegion]] -- generated region lists
        """
        step = int(step_style) if step_style > 1 else int(size * step_style)
        if step < size:
            return self.__build_overlapping(size, step)
        return self.__build_non_overlapping(size, step)



#!/usr/bin/python3

from Bio.SeqIO.FastaIO import SimpleFastaParser
import numpy as np
import pandas as pd
import os
import glob
import sys
from tqdm import tqdm 
from tabulate import tabulate
import statistics as stat
import shutil



def select_probe(currentfolder:os.PathLike = './data/')->None:
    # parameters
    cutoff_cost = 1e5       # max total cost of a probe. Exclude probe if even one oligo has a prohibitive cost
    cutoff_d_pc = 10        # max distance between 2 consecutive oligos, as a % of the total probe length
    if (len(sys.argv)==2):
        cutoff_d = sys.argv[1]
    else:
        cutoff_d = 500          # max distance between 2 consecutive oligos, in nucleotides

    # retrieve probe queries
    selectedfolder = currentfolder + 'selected_probes/'
    try:
        os.mkdir(selectedfolder)
    except FileExistsError:
        pass
    # identify probe files
    pattern = currentfolder+"probe_candidates/**/probe_*.tsv"   #probelet
    #pattern = "data/**/probe*/**oligos.tsv"    #ifpd2 query
    #pattern = "data/**/*.best_probe.tsv"   #jupyternb
    filenames = glob.glob(pattern,recursive=True)

    roilist = currentfolder+'rois/all_regions.tsv'
    rdroi = pd.read_csv(roilist,sep="\t",header=0)


    probelist = pd.DataFrame(columns=['fullpath', 'folder', 'probe_set', 'roi', 'chr', 'probe_start', 'probe_end', \
            'nOligos', 'pw', 'probe_size', 'region_size', 'coverage', \
            'centrality', 'd_max_pcregion', 'd_max_pcprobe',\
            'd_mean', 'd_min', 'd_max', 'd_std', \
            'tm_range', 'tm_mean', 'tm_std', \
            'gc_range', 'gc_mean', 'gc_std', \
            'mean_closestMM', 'min_closestMM', 'max_closestMM', 'std_closestMM', \
            'mean_cumulMM', 'min_cumulMM', 'max_cumulMM', 'std_cumulMM', \
            'mean_oligo_cost', 'min_oligo_cost', 'max_oligo_cost', 'std_oligo_cost' ,\
            'sum_inv_cost'])
            #'mean_sumMM', 'min_sumMM', 'max_sumMM', 'std_sumMM'   ])
        

    # read probe files
    for file in tqdm(filenames, "Processing probe candidates"):
        rd = pd.read_csv(file,sep="\t")
        last = len(probelist)+1
        
        # find coordinates of the corresponding roi
        current_roi = rd.name[1][4:]
        for k in range(len(rdroi)):
            if str(rdroi.window_id[k]) == current_roi:
                current = k

        roi_start = rdroi.Window_start[current]
        roi_end = rdroi.Window_end[current]
        roi_center = ((roi_start+roi_end)/2)-roi_start  #adjusted roi center

    # calculate summary measures and add as last row

        inv_cost = np.divide(np.ones(len(rd.name.tolist())),rd.oligo_cost)

    # replace off_target_no by min_mismatch for new queries
        probe_start = min(rd.start[1:])
        probe_end = max(rd.end[1:])
        probe_center = ((probe_start+probe_end)/2)-roi_start    #adjusted probe center

        filesplit = file.split("/")

        probelist.loc[last] = [file, filesplit[3], filesplit[4], rd.name[1], rd.chromosome[1], probe_start, probe_end, \
            len(rd), float(filesplit[4][-8:-4]), probe_end-probe_start+1, int(roi_end)-int(roi_start)+1, (probe_end-probe_start+1)/(int(roi_end)-int(roi_start)+1), \
            min(probe_center/roi_center,2-(probe_center/roi_center)), 100*(rd.start-rd.end.shift()).max()/(int(roi_end)-int(roi_start)+1), 100*(rd.start-rd.end.shift()).max()/(int(probe_end)-int(probe_start)+1),\
            (rd.start-rd.end.shift()).mean(), (rd.start-rd.end.shift()).min(), (rd.start-rd.end.shift()).max(), (rd.start-rd.end.shift()).std(), \
            max(rd.Tm)-min(rd.Tm), stat.mean(rd.Tm), stat.stdev(rd.Tm), \
            max(rd.gc_content)-min(rd.gc_content), stat.mean(rd.gc_content), stat.stdev(rd.gc_content), \
            stat.mean(rd.off_target_no), min(rd.off_target_no), max(rd.off_target_no), stat.stdev(rd.off_target_no), \
            stat.mean(rd.off_target_sum), min(rd.off_target_sum), max(rd.off_target_sum), stat.stdev(rd.off_target_sum), \
            stat.mean(rd.oligo_cost), min(rd.oligo_cost), max(rd.oligo_cost), stat.stdev(rd.oligo_cost), \
            inv_cost.sum()]
        # stat.mean(rd.off_target_sum), min(rd.off_target_sum), max(rd.off_target_sum), stat.stdev(rd.off_target_sum)]  
    
    probelist.loc[:,'fullpath'] = probelist.loc[:,'fullpath'].astype(str)

    # one selected probe per ROI
    for roi in tqdm(pd.unique(probelist.roi),"Selecting optimal probes..."):
        roiprobelist = probelist[(probelist.max_oligo_cost < cutoff_cost) & (probelist.d_max_pcprobe < cutoff_d_pc) & (probelist.d_max < cutoff_d) & (probelist.roi == roi)]
        filteredroilist = roiprobelist[roiprobelist.nOligos == roiprobelist.nOligos.max()]
        selectedprobeindex = filteredroilist[['pw']].idxmin()
        shutil.copy2(filteredroilist.loc[selectedprobeindex].fullpath.tolist()[0],selectedfolder)

    return

if __name__ == "__main__":
    select_probe()


"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

import click  # type: ignore
from ifpd2.const import CONTEXT_SETTINGS
from ifpd2.database import DataBase
import logging


@click.command(
    name="info", context_settings=CONTEXT_SETTINGS, help="Show INPUT database details."
)
@click.argument("input_paths", metavar="INPUT", nargs=1, type=click.Path(exists=True))
def main(input_paths: str) -> None:
    DB = DataBase(input_paths)
    DB.log_details()
    logging.info("")
    logging.info("That's all! :smiley:")



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

import logging
import numpy as np  # type: ignore
import sys
from typing import Callable


def ert_type(x, stype, label):
    if not isinstance(x, stype):
        raise AssertionError(f"{label} should be {stype}, {type(x)} instead")


def ert_multiTypes(x, types, label):
    cond = any(isinstance(x, t) for t in types)
    if not cond:
        raise AssertionError(f"{label} should be one of {types}, {type(x)} instead")


def ert_nonNeg(x, label, include_zero=False):
    if not include_zero:
        if x <= 0:
            raise AssertionError(f"{label} should be greater than 0")
    elif x < 0:
        raise AssertionError(f"{label} should be greater than or equal to 0")


def ert_inInterv(x, vmin, vmax, label, leftClose=False, rightClose=True):
    if leftClose:
        if rightClose:
            if x < vmin or x > vmax:
                raise AssertionError(f"expected {vmin}<={label}<={vmax}")
        elif x < vmin or x >= vmax:
            raise AssertionError(f"expected {vmin}<={label}<{vmax}")
    elif rightClose:
        if x <= vmin or x > vmax:
            raise AssertionError(f"expected {vmin}<{label}<={vmax}")
    elif x <= vmin or x >= vmax:
        raise AssertionError(f"expected {vmin}<{label}<{vmax}")


def ert_in_dtype(x, dtype):
    if dtype.startswith("f"):
        if x > np.finfo(dtype).max:
            raise AssertionError(
                " ".join(
                    [
                        "expected to be lower than {dtype} max:",
                        f"{x} < {np.finfo(dtype).max}",
                    ]
                )
            )
    elif dtype.startswith("u") or dtype.startswith("i"):
        if x > np.iinfo(dtype).max:
            raise AssertionError(
                " ".join(
                    [
                        "expected to be lower than {dtype} max:",
                        f"{x} < {np.iinfo(dtype).max}",
                    ]
                )
            )
    else:
        logging.warning(f"assert not implemented for dtype '{dtype}'")


def enable_rich_assert(fun: Callable) -> Callable:
    def wrapper(*args, **kwargs):
        try:
            return fun(*args, **kwargs)
        except AssertionError as e:
            logging.exception(e)
            sys.exit()

    return wrapper



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

import itertools
import logging
import numpy as np  # type: ignore
import os
import pandas as pd  # type: ignore
from rich.progress import track  # type: ignore
import shutil
from typing import List

from ifpd2.probe import OligoProbe


class OligoProbeSet:
    """Set of non-overlapping probes built from windows."""

    def __init__(self, probe_list):
        super(OligoProbeSet, self).__init__()
        self.probe_list = probe_list

    @property
    def probe_list(self):
        return self._probe_list

    @probe_list.setter
    def probe_list(self, probe_list):
        if not all(isinstance(p, OligoProbe) for p in probe_list):
            raise AssertionError
        self._probe_list = sorted(probe_list, key=lambda p: p.range[0])
        self._probe_tm_ranges = [p.tm_range for p in self.probe_list]
        self._tm_range = (
            np.min([t[0] for t in self._probe_tm_ranges]),
            np.max([t[1] for t in self._probe_tm_ranges]),
        )
        self._sizes = [p.size for p in self.probe_list]
        self._spreads = [p.spread for p in self.probe_list]
        self._probe_ranges = [p.range for p in self.probe_list]
        probe_starts = np.array([r[0] for r in self.probe_ranges])
        probe_ends = np.array([r[1] for r in self.probe_ranges])
        self._range = (probe_starts.min(), probe_ends.max())
        self._ds = 0 if len(probe_list) == 1 else probe_starts[1:] - probe_ends[:-1]
        self._d_mean = np.mean(self.ds)
        self._d_range = (np.min(self.ds), np.max(self.ds))
        tm = [p.data["Tm"].tolist() for p in self.probe_list]
        self._oligo_tm = list(itertools.chain(*tm))

    @property
    def range(self):
        return self._range

    @property
    def probe_ranges(self):
        return self._probe_ranges

    @property
    def ds(self):
        return self._ds

    @property
    def d_mean(self):
        return self._d_mean

    @property
    def d_range(self):
        return self._d_range

    @property
    def tm_range(self):
        return self._tm_range

    @property
    def probe_tm_ranges(self):
        return self._probe_tm_ranges

    @property
    def oligo_tm(self):
        return self._oligo_tm

    @property
    def sizes(self):
        return self._sizes

    @property
    def spreads(self):
        return self._spreads

    @property
    def featDF(self):
        mean_score = (1 / 3) * sum(
            [
                np.std(self.sizes) / np.mean(self.sizes),
                np.std(self.spreads) / np.mean(self.spreads),
                np.std(self.oligo_tm) / np.mean(self.oligo_tm),
            ]
        )
        df = pd.DataFrame(
            [
                self.range[0],
                self.range[1],
                len(self.probe_list),
                np.diff(self.range)[0],
                np.mean(self.sizes),
                np.std(self.sizes),
                np.min(self.spreads),
                np.max(self.spreads),
                np.mean(self.spreads),
                np.std(self.spreads),
                self.d_range[0],
                self.d_range[1],
                self.d_mean,
                np.std(self.ds),
                np.diff(self.tm_range)[0],
                np.mean(self.oligo_tm),
                np.std(self.oligo_tm),
                mean_score,
            ]
        ).transpose()
        df.columns = [
            "start",
            "end",
            "nProbes",
            "size",
            "size_mean",
            "size_std",
            "spread_min",
            "spread_max",
            "spread_mean",
            "spread_std",
            "d_min",
            "d_max",
            "d_mean",
            "d_std",
            "tm_range",
            "tm_mean",
            "tm_std",
            "score",
        ]
        return df

    def export(self, path):
        if os.path.isfile(path):
            raise AssertionError
        if os.path.isdir(path):
            shutil.rmtree(path)
        os.mkdir(path)

        self.featDF.to_csv(os.path.join(path, "set.tsv"), "\t", index=False)

        pd.concat([p.featDF for p in self.probe_list]).reset_index(drop=True).to_csv(
            os.path.join(path, "probes.tsv"), "\t"
        )

        with open(os.path.join(path, "set.bed"), "w+") as BH:
            for pi in range(len(self.probe_list)):
                probe = self.probe_list[pi]
                probe.export(os.path.join(path, f"probe_{pi}"))
                for i in range(probe.data.shape[0]):
                    oligo = probe.data.iloc[i]
                    BH.write(
                        "".join(
                            [
                                f">probe_{pi}:{oligo['name']}",
                                f":{oligo['chromosome']}:",
                                f"{oligo['start']}-{oligo['end']}\n",
                            ]
                        )
                    )
                    BH.write(f"{probe.data.iloc[i]['sequence']}\n")


class OligoProbeSetBuilder:
    """Class to build OligoProbeSet objects."""

    probe_set_list: List = []

    def __init__(self, out_path):
        super(OligoProbeSetBuilder, self).__init__()
        if os.path.isfile(out_path):
            raise AssertionError
        self.out_path = out_path
        if os.path.isdir(out_path):
            shutil.rmtree(out_path)
        os.mkdir(out_path)

    @staticmethod
    def __build_probe_set_list(window_list, i):
        probe_set_list = [(p,) for p in window_list[i]]

        for w in window_list[(i + 1) :]:
            if len(w) == 0:
                continue

            current_probe_set_list = list(tuple(probe_set_list))
            probe_set_list = []

            for probe in w:
                for probe_set in current_probe_set_list:
                    current_probe_set = list(probe_set)
                    current_probe_set.append(probe)
                    probe_set_list.append(current_probe_set)

        return probe_set_list

    def build(self, probe_candidates):
        for (wSet, window_list) in probe_candidates.items():
            window_list = list(window_list.values())

            non_empty_ids = [
                i for i in range(len(window_list)) if len(window_list[i]) != 0
            ]

            if not non_empty_ids:
                logging.warning(f"No probe candidates found, dropped. [ws{wSet+1}]")
                continue
            i = non_empty_ids[0]

            probe_set_list = self.__build_probe_set_list(window_list, i)

            self.probe_set_list.extend(probe_set_list)
            logging.info(
                "".join(
                    [
                        f"Built {len(probe_set_list)} probe set candidates ",
                        f"from window set #{wSet+1}",
                    ]
                )
            )

        self.probe_set_list = [OligoProbeSet(ps) for ps in self.probe_set_list]
        logging.info(
            f"Built {len(self.probe_set_list)} " + "probe set candidates in total."
        )

        if self.probe_set_list:
            pd.concat([ps.featDF for ps in self.probe_set_list]).reset_index(
                drop=True
            ).to_csv(os.path.join(self.out_path, "probe_sets.tsv"), "\t")

    def export(self):
        logging.info("Exporting probe sets...")
        for psi in track(range(len(self.probe_set_list)), description="Probe set"):
            probe_set = self.probe_set_list[psi]
            probe_set.export(os.path.join(self.out_path, f"probe_set_{psi}"))



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

import click  # type: ignore
from ifpd2.const import CONTEXT_SETTINGS
from ifpd2.dataclasses import Folder, GenomicRegion, FreeEnergyInterval
from ifpd2.dataclasses import NonNegativeInteger, PositiveInteger
from ifpd2.dataclasses import PositiveFloat
from ifpd2.dataclasses import NonNegativeIntInterval
from ifpd2.dataclasses import QueryFocus, QueryWindow
from ifpd2.walker import Walker
from ifpd2.logging import add_log_file_handler
from ifpd2.probe import OligoProbeBuilder
from ifpd2.probe_set import OligoProbeSetBuilder
import logging
from os.path import isdir, isfile, join as path_join
from os import mkdir
from typing import Optional, Tuple


@click.command(
    name="query",
    context_settings=CONTEXT_SETTINGS,
    help="""Design FISH probes on a database chromosome.
    When a --region is not specified, the whole chromosome is used.
    Use -M for the number of probes, and -N for the number of oligos per probe.
    Use -D to specify the minimum distance between consecutive oligos in a probe.

    Use --single to access single-probe design. Same as "-X 1 -w 1", this option
    overrides -X and -w, and ignores -W, if specified.""",
)
@click.argument("db_folder", type=click.Path(exists=True))
@click.argument("chromosome", type=click.STRING)
@click.argument("output_path", metavar="OUTPUT", type=click.Path(exists=False))
@click.option(
    "--region",
    type=click.INT,
    nargs=2,
    default=(0, -1),
    help="Space-separate region start and end location.",
)
@click.option("-M", "--probes", type=click.INT, help="Design M probes.")
@click.option(
    "-N",
    "--oligos",
    type=click.INT,
    default=48,
    help="N oligos per probe. Default: 48",
)
@click.option(
    "-D",
    "--dist",
    type=click.INT,
    default=2,
    help="Consecutive oligo distance. Default: 2",
)
@click.option("--single", is_flag=True, help="Easy access to single probe design.")
@click.option("-W", "--window-size", type=click.INT)
@click.option(
    "-w",
    "--window-shift",
    type=click.FLOAT,
    default=0.1,
    help="Shift as window fraction. Default: 0.1",
)
@click.option(
    "-R",
    "--focus-size",
    type=click.FLOAT,
    default=8e3,
    help="""\b
    Focus size in nt or window fraction.
    Default: 8000""",
)
@click.option(
    "-r",
    "--focus-step",
    type=click.FLOAT,
    default=1e3,
    help="""\b
    Focus step in nt or focus fraction.
    Default: 1000""",
)
@click.option(
    "-F",
    "--off-targets",
    type=click.INT,
    nargs=2,
    default=(0, 99),
    help="""Extremities of acceptable off-target range.
    Default: [0, 99]""",
)
@click.option(
    "-G",
    "--free-energy",
    type=click.FLOAT,
    nargs=2,
    default=(0.0, 0.5),
    help="""Extremities of acceptable 2nd structure dG range,
    in absolute kcal/mol or hybridization dG fraction.
    Default: [0.0, 0.5]""",
)
@click.option(
    "-o",
    "--oligo-score-step",
    type=click.FLOAT,
    default=0.1,
    help="Oligo score relaxation step. Default: 0.1",
)
@click.option(
    "-t",
    "--melting-half-width",
    type=click.FLOAT,
    default=10.0,
    help="Melting range half-width. Default: 10.0",
)
@click.option(
    "-P",
    "--probe-size",
    type=click.INT,
    default=1e4,
    help="Max probe size in nt. Default: 10000",
)
@click.option(
    "-H",
    "--gap-size",
    type=click.FLOAT,
    default=0.1,
    help="""\b
    Max gap size as probe size fraction.
    Default: 0.1""",
)
@click.option(
    "-I",
    "--oligo-intersection",
    type=click.FLOAT,
    default=0.5,
    help="""Probe intersection threshold as shared oligo fraction.
    Default: .5""",
)
@click.option(
    "-k",
    "--oligo-length",
    type=click.INT,
)
@click.option(
    "--threads",
    type=click.INT,
    default=1,
    help="Threads for parallelization. Default: 1",
)
def main(
    db_folder: str,
    chromosome: str,
    output_path: str,
    region: Tuple[int, int],
    probes: Optional[int],
    oligos: int,
    dist: int,
    single: bool,
    window_size: Optional[int],
    window_shift: float,
    focus_size: float,
    focus_step: float,
    off_targets: Tuple[int, int],
    free_energy: Tuple[float, float],
    oligo_score_step: float,
    melting_half_width: float,
    probe_size: int,
    gap_size: float,
    oligo_intersection: float,
    oligo_length: Optional[int],
    threads: int,
) -> None:
    settings = QuerySettings(db_folder, output_path)

    if single:
        probes = 1
        window_shift = 1.0
        window_size = None

    if probes is None and window_size is None:
        raise AssertionError
    if region[1] <= 0:
        probes = None
        logging.info(
            " ".join(
                [
                    "Cannot design a specific number of probes for queries on",
                    "a whole chromosome/feature.",
                ]
            )
        )

    assert_reusable(settings.output_path)

    if not isdir(settings.output_path):
        mkdir(settings.output_path)
    add_log_file_handler(f"{settings.output_path}/ifpd2-main.log")

    opb = OligoProbeBuilder()
    opb.N = PositiveInteger(oligos).n
    opb.D = NonNegativeInteger(dist).n
    opb.Tr = PositiveFloat(melting_half_width).n
    opb.Ps = PositiveInteger(probe_size).n
    opb.Ph = PositiveFloat(gap_size, 1).n
    opb.Po = PositiveFloat(oligo_intersection, 1).n
    opb.k = oligo_length
    opb.F = NonNegativeIntInterval(*off_targets).astuple()
    opb.Gs = FreeEnergyInterval(*free_energy).astuple()
    opb.Ot = PositiveFloat(oligo_score_step, 1).n

    opsb = OligoProbeSetBuilder(path_join(settings.output_path, "probe_sets"))
    logging.info(opb.get_prologue())

    ow = Walker(settings.db_folder_path)
    ow.C = chromosome
    genomic_region = GenomicRegion(*region)
    ow.S = genomic_region.start
    ow.E = max(genomic_region.end, genomic_region.start)
    ow.X = 1 if probes is None else PositiveInteger(probes).n
    window = QueryWindow(window_size, window_shift)
    ow.Ws = window.size
    ow.Wh = window.shift
    focus = QueryFocus(focus_size, focus_step)
    ow.Rs = focus.size
    ow.Rt = focus.step
    ow.out_path = settings.output_path
    ow.reuse = False
    ow.threads = PositiveInteger(threads).n

    ow.start(
        start_from_nt=genomic_region.start,
        end_at_nt=genomic_region.end,
        N=opb.N,
        opb=opb,
        cfr_step=ow.Rt,
    )

    opsb.build(ow.walk_results)
    opsb.export()

    logging.info("Done. :thumbs_up: :smiley:")
    logging.shutdown()


class QuerySettings:
    _db_folder: Folder
    _output_path: Folder

    def __init__(self, db_folder: str, output_path: str):
        super(QuerySettings, self).__init__()
        self._db_folder = Folder(db_folder, True)
        self._output_path = Folder(output_path, False)

    @property
    def db_folder_path(self) -> str:
        return self._db_folder.path

    @property
    def output_path(self) -> str:
        return self._output_path.path

    @property
    def reusable(self) -> bool:
        return False


def assert_reusable(output_path: str):
    assert_msg = " ".join(
        [
            "output path should NOT direct towards an existing directory",
            "or file. Use '--reuse' to load previous results.",
            f"Provided path '{output_path}'",
        ]
    )
    if isfile(output_path):
        raise AssertionError(assert_msg + " leads to a file")
    if isdir(output_path):
        raise AssertionError(assert_msg + " leads to a directory.")



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

from ifpd2 import asserts, dataclasses, fasta, io
from ifpd2 import walker, walker2
from ifpd2 import chromosome, database, region
from ifpd2 import oligo, probe, probe_set

from importlib.metadata import version, PackageNotFoundError
from typing import List

try:
    __version__ = version(__name__)
except PackageNotFoundError:
    pass

__all__ = [
    "__version__",
    "asserts",
    "dataclasses",
    "fasta",
    "io",
    "walker",
    "walker2",
    "chromosome",
    "database",
    "region",
    "oligo",
    "probe",
    "probe_set",
]
__path__: List[str]



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

from Bio.SeqIO.FastaIO import SimpleFastaParser  # type: ignore
from os.path import isfile
from tqdm import tqdm  # type: ignore
from typing import List, Tuple


def extract_kmers(input_path: str, kmer_size: int) -> List[Tuple[str, str]]:
    if not isfile(input_path):
        raise AssertionError
    with open(input_path) as FH:
        oligos_list: List[Tuple[str, str]] = []
        for record_header, record_sequence in tqdm(
            SimpleFastaParser(FH), desc="Parsing region", leave=False
        ):
            for i in tqdm(
                range(len(record_sequence) - kmer_size + 1),
                desc="Extracting oligos",
                leave=False,
            ):
                oligo_sequence = str(record_sequence)[slice(i, i + kmer_size)]
                if "N" in oligo_sequence:
                    continue
                oligos_list.append(
                    (
                        f"{record_header}|{i+1}:{i+kmer_size+1}",
                        oligo_sequence,
                    )
                )
    return oligos_list



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

import click  # type: ignore
from ifpd2.const import CONTEXT_SETTINGS
from ifpd2.dataclasses import Folder
from ifpd2.dataclasses import PositiveInteger
from ifpd2.dataclasses import QueryWindow
from ifpd2.logging import add_log_file_handler
from ifpd2.database import DataBase
from ifpd2.region import GenomicRegionBuilder
from ifpd2.walker2 import ChromosomeWalker
import logging
from os.path import isdir, isfile
from os import mkdir
from typing import Optional, Tuple


@click.command(
    name="query2",
    context_settings=CONTEXT_SETTINGS,
    help="""Design FISH probes on a database chromosome.
    When a --region is not specified, the whole chromosome is used.
    Use -M for the number of probes, and -N for the number of oligos per probe.
    Use -D to specify the minimum distance between consecutive oligos in a probe.

    Use --single to access single-probe design. Same as "-X 1 -w 1", this option
    overrides -X and -w, and ignores -W, if specified.""",
)
@click.argument("db_folder", type=click.Path(exists=True))
@click.argument("chromosome", type=click.STRING)
@click.argument("output_path", metavar="OUTPUT", type=click.Path(exists=False))
@click.option(
    "--region",
    type=click.INT,
    nargs=2,
    default=(0, -1),
    help="Space-separate region start and end location.",
)
@click.option("-M", "--probes", type=click.INT, help="Design M probes.")
@click.option(
    "-N",
    "--oligos",
    type=click.INT,
    default=48,
    help="N oligos per probe. Default: 48",
)
@click.option(
    "-D",
    "--dist",
    type=click.INT,
    default=2,
    help="Consecutive oligo distance. Default: 2",
)
@click.option("--single", is_flag=True, help="Easy access to single probe design.")
@click.option("-W", "--window-size", type=click.INT)
@click.option(
    "-w",
    "--window-shift",
    type=click.FLOAT,
    default=0.1,
    help="Shift as window fraction. Default: 0.1",
)
@click.option(
    "-R",
    "--focus-size",
    type=click.FLOAT,
    default=8e3,
    help="""\b
    Focus size in nt or window fraction.
    Default: 8000""",
)
@click.option(
    "-r",
    "--focus-step",
    type=click.FLOAT,
    default=1e3,
    help="""\b
    Focus step in nt or focus fraction.
    Default: 1000""",
)
@click.option(
    "-F",
    "--off-targets",
    type=click.INT,
    nargs=2,
    default=(0, 99),
    help="""Extremities of acceptable off-target range.
    Default: [0, 99]""",
)
@click.option(
    "-G",
    "--free-energy",
    type=click.FLOAT,
    nargs=2,
    default=(0.0, 0.5),
    help="""Extremities of acceptable 2nd structure dG range,
    in absolute kcal/mol or hybridization dG fraction.
    Default: [0.0, 0.5]""",
)
@click.option(
    "-o",
    "--oligo-score-step",
    type=click.FLOAT,
    default=0.1,
    help="Oligo score relaxation step. Default: 0.1",
)
@click.option(
    "-t",
    "--melting-half-width",
    type=click.FLOAT,
    default=10.0,
    help="Melting range half-width. Default: 10.0",
)
@click.option(
    "-P",
    "--probe-size",
    type=click.INT,
    default=1e4,
    help="Max probe size in nt. Default: 10000",
)
@click.option(
    "-H",
    "--gap-size",
    type=click.FLOAT,
    default=0.1,
    help="""\b
    Max gap size as probe size fraction.
    Default: 0.1""",
)
@click.option(
    "-I",
    "--oligo-intersection",
    type=click.FLOAT,
    default=0.5,
    help="""Probe intersection threshold as shared oligo fraction.
    Default: .5""",
)
@click.option(
    "-k",
    "--oligo-length",
    type=click.INT,
)
@click.option(
    "--threads",
    type=click.INT,
    default=1,
    help="Threads for parallelization. Default: 1",
)
def main(
    db_folder: str,
    chromosome: str,
    output_path: str,
    region: Tuple[int, int],
    probes: Optional[int],
    oligos: int,
    dist: int,
    single: bool,
    window_size: Optional[int],
    window_shift: Optional[float],
    focus_size: float,
    focus_step: float,
    off_targets: Tuple[int, int],
    free_energy: Tuple[float, float],
    oligo_score_step: float,
    melting_half_width: float,
    probe_size: int,
    gap_size: float,
    oligo_intersection: float,
    oligo_length: Optional[int],
    threads: int,
) -> None:
    settings = QuerySettings(db_folder, output_path)

    probes, window_size, window_shift = check_input(
        single, probes, window_size, window_shift
    )

    if not isdir(settings.output_path):
        mkdir(settings.output_path)
    add_log_file_handler(f"{settings.output_path}/ifpd2-main.log")

    DB = DataBase(settings.db_folder_path)

    RB = GenomicRegionBuilder(DB.get_chromosome(chromosome.encode()))
    if probes is not None:
        region_set_list = RB.build_by_number(PositiveInteger(probes).n)
    else:
        window_size, window_shift = QueryWindow(window_size, window_shift).astuple()
        if window_size is None or window_shift is None:
            raise AssertionError
        region_set_list = RB.build_by_size(window_size, window_shift)

    walker = ChromosomeWalker(DB, chromosome.encode())
    walker.walk_multiple_regions(region_set_list)

    # print((DB, RB, walker, region_set_list))

    logging.info("Done. :thumbs_up: :smiley:")
    logging.shutdown()


def check_input(
    single: bool,
    probes: Optional[int],
    window_size: Optional[int],
    window_shift: Optional[float],
) -> Tuple[Optional[int], Optional[int], Optional[float]]:
    if single:
        probes = 1
        window_shift = None
        window_size = None

    if probes is None and window_size is None:
        raise AssertionError
    if probes is not None and window_size is not None:
        logging.warning("cannot combine -X and -W. Using -X.")
        window_size = None
        window_shift = None
    if window_size is not None and window_shift is None:
        window_shift = 1

    return (probes, window_size, window_shift)


class QuerySettings:
    _db_folder: Folder
    _output_path: Folder

    def __init__(self, db_folder: str, output_path: str):
        super(QuerySettings, self).__init__()
        self._db_folder = Folder(db_folder, True)
        self._output_path = Folder(output_path, False)

    @property
    def db_folder_path(self) -> str:
        return self._db_folder.path

    @property
    def output_path(self) -> str:
        return self._output_path.path

    @property
    def reusable(self) -> bool:
        return False


def assert_reusable(output_path: str):
    assert_msg = " ".join(
        [
            "output path should NOT direct towards an existing directory",
            "or file. Use '--reuse' to load previous results.",
            f"Provided path '{output_path}'",
        ]
    )
    if isfile(output_path):
        raise AssertionError(assert_msg + " leads to a file")
    if isdir(output_path):
        raise AssertionError(assert_msg + " leads to a directory.")
